#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <inttypes.h>
#include <time.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/utsname.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/uio.h>

#include <xs.h>
#include "xennerctrl.h"
#include <xen/io/xenbus.h>

#include "blkif.h"
#include "shared.h"
#include "list.h"
#include "hexdump.h"
#include "daemon.h"
#include "xenbackd.h"
#include "evtchnd.h"

/* ------------------------------------------------------------- */

#define BE_ROOT "/local/domain/0/backend/blkbackd"

enum blkif_state {
    BLKIF_UNKNOWN = 0,
    BLKIF_PROBED,
    BLKIF_INITIALISED,
    BLKIF_CONNECTED,
    BLKIF_DISCONNECTED,
};

struct blkdev {
    struct xendev       xendev;  /* must be first */
    char                *params;
    char                *mode;
    int                 file;
    size_t              file_blk;
    size_t              file_size;
    int                 protocol;
    blkif_back_rings_t  rings;
    int                 more_work;
    enum blkif_state    state;
};

static int termsig;

/* ------------------------------------------------------------- */

static void catchsig(int sig)
{
    termsig = sig;
}

/* ------------------------------------------------------------- */

static void blk_setup_backend(struct blkdev *blkdev)
{
    int mode, info = 0;
    
    if (blkdev->state != BLKIF_PROBED)
	return;
    
    /* read xenstore entries */
    if (NULL == blkdev->params)
	blkdev->params = read_be_str(&blkdev->xendev, "params");
    if (NULL == blkdev->mode)
	blkdev->mode = read_be_str(&blkdev->xendev, "mode");

    /* do we have all we need? */
    if (NULL == blkdev->params ||
	NULL == blkdev->mode)
	return;

    /* try opening file */
    if (0 == strcmp(blkdev->mode, "w")) {
	mode = O_RDWR;
    } else {
	mode = O_RDONLY;
	info |= VDISK_READONLY;
    }
    blkdev->file = open(blkdev->params, mode);
    if (-1 == blkdev->file) {
	d1printf("%s: open %s: %s\n", __FUNCTION__,
		 blkdev->params, strerror(errno));
	return;
    }
    
    /* fill info */
    blkdev->file_blk  = 512;
    blkdev->file_size = lseek(blkdev->file, 0, SEEK_END);
    write_be_int(&blkdev->xendev, "feature-barrier", 1);
    write_be_int(&blkdev->xendev, "info",            info);
    write_be_int(&blkdev->xendev, "sector-size",     blkdev->file_blk);
    write_be_int(&blkdev->xendev, "sectors",
		 blkdev->file_size / blkdev->file_blk);

    change_state_xendev(&blkdev->xendev, XenbusStateInitWait);
    blkdev->state = BLKIF_INITIALISED;
}

static void blk_connect_frontend(struct blkdev *blkdev)
{
    void *sring;
    
    if (blkdev->state != BLKIF_INITIALISED)
	return;

    blkdev->xendev.ring_ref = read_fe_int(&blkdev->xendev, "ring-ref");
    blkdev->xendev.remote_port = read_fe_int(&blkdev->xendev, "event-channel");
    blkdev->xendev.protocol = read_fe_str(&blkdev->xendev, "protocol");
    if (0 == blkdev->xendev.ring_ref ||
	0 == blkdev->xendev.remote_port ||
	NULL == blkdev->xendev.protocol)
	return;

    blkdev->protocol = BLKIF_PROTOCOL_NATIVE;
    if (0 == strcmp(blkdev->xendev.protocol, XEN_IO_PROTO_ABI_X86_32))
	blkdev->protocol = BLKIF_PROTOCOL_X86_32;
    if (0 == strcmp(blkdev->xendev.protocol, XEN_IO_PROTO_ABI_X86_64))
	blkdev->protocol = BLKIF_PROTOCOL_X86_64;

    if (vmcore_map(&blkdev->xendev) < 0)
	return;
    sring = vmcore_ref2ptr(&blkdev->xendev, blkdev->xendev.ring_ref);
    if (!sring)
	return;
    switch (blkdev->protocol) {
    case BLKIF_PROTOCOL_NATIVE:
    {
	blkif_sring_t *sring_native = sring;
	BACK_RING_INIT(&blkdev->rings.native, sring_native, PAGE_SIZE);
	break;
    }
    case BLKIF_PROTOCOL_X86_32:
    {
	blkif_x86_32_sring_t *sring_x86_32 = sring;
	BACK_RING_INIT(&blkdev->rings.x86_32, sring_x86_32, PAGE_SIZE);
	break;
    }
    case BLKIF_PROTOCOL_X86_64:
    {
	blkif_x86_64_sring_t *sring_x86_64 = sring;
	BACK_RING_INIT(&blkdev->rings.x86_64, sring_x86_64, PAGE_SIZE);
	break;
    }
    }

    blkdev->xendev.local_port = xenner_evtchn_bind_interdomain
	(evtchnd, blkdev->xendev.dom, blkdev->xendev.remote_port);

    d1printf("%s: ok: proto %s, ring-ref %d, "
	     "remote port %d, local port %d\n", __FUNCTION__,
	     blkdev->xendev.protocol, blkdev->xendev.ring_ref,
	     blkdev->xendev.remote_port, blkdev->xendev.local_port);

    change_state_xendev(&blkdev->xendev, XenbusStateConnected);
    blkdev->state = BLKIF_CONNECTED;
}

static void blk_disconnect(struct blkdev *blkdev, enum xenbus_state state)
{
    if (-1 != blkdev->file) {
	close(blkdev->file);
	blkdev->file = -1;
    }
    if (blkdev->xendev.vmcore)
	vmcore_unmap(&blkdev->xendev);
    if (blkdev->xendev.local_port) {
	xenner_evtchn_unbind(evtchnd, blkdev->xendev.local_port);
	blkdev->xendev.local_port = 0;
    }

    if (state)
	change_state_xendev(&blkdev->xendev, state);
    blkdev->state = BLKIF_DISCONNECTED;
}

static void blk_send_response(struct blkdev *blkdev, uint64_t id,
			      unsigned short op, int st)
{
    int               send_notify   = 0;
    int               have_requests = 0;
    blkif_response_t  resp;
    void              *dst;
    
    resp.id        = id;
    resp.operation = op;
    resp.status    = st;
    
    /* Place on the response ring for the relevant domain. */
    switch (blkdev->protocol) {
    case BLKIF_PROTOCOL_NATIVE:
	dst = RING_GET_RESPONSE(&blkdev->rings.native, blkdev->rings.native.rsp_prod_pvt);
	break;
    case BLKIF_PROTOCOL_X86_32:
	dst = RING_GET_RESPONSE(&blkdev->rings.x86_32, blkdev->rings.x86_32.rsp_prod_pvt);
	break;
    case BLKIF_PROTOCOL_X86_64:
	dst = RING_GET_RESPONSE(&blkdev->rings.x86_64, blkdev->rings.x86_64.rsp_prod_pvt);
	break;
    }
    memcpy(dst, &resp, sizeof(resp));
    blkdev->rings.common.rsp_prod_pvt++;

    RING_PUSH_RESPONSES_AND_CHECK_NOTIFY(&blkdev->rings.common, send_notify);
    if (blkdev->rings.common.rsp_prod_pvt == blkdev->rings.common.req_cons) {
	/*
	 * Tail check for pending requests. Allows frontend to avoid
	 * notifications if requests are already in flight (lower
	 * overheads and promotes batching).
	 */
	RING_FINAL_CHECK_FOR_REQUESTS(&blkdev->rings.common, have_requests);
    } else if (RING_HAS_UNCONSUMED_REQUESTS(&blkdev->rings.common)) {
	have_requests = 1;
    }

    d2printf("%s: send_notify %d, have_requests %d\n",
	     __FUNCTION__, send_notify, have_requests);

    if (have_requests)
	blkdev->more_work++;
    if (send_notify)
	xenner_evtchn_notify(evtchnd, blkdev->xendev.local_port);
}

static void blk_handle_request(struct blkdev *blkdev, blkif_request_t *req)
{
    struct iovec vec[BLKIF_MAX_SEGMENTS_PER_REQUEST];
    void *page, *mem;
    off_t start, end;
    size_t len;
    int i;
    
    d2printf("%s: op %d, nr %d, handle %d, id %" PRId64 ", sector %" PRId64 "\n",
	     __FUNCTION__, req->operation, req->nr_segments,
	     req->handle, req->id, req->sector_number);
    if (BLKIF_OP_READ != req->operation && blkdev->mode[0] != 'w')
	goto err; /* write request for r/o disk */

    start = end = req->sector_number * blkdev->file_blk;
    for (i = 0; i < req->nr_segments; i++) {
	d2printf("%s:   %d: gref %d sects %d-%d\n", __FUNCTION__,
		 i, req->seg[i].gref, req->seg[i].first_sect, req->seg[i].last_sect);

	page = vmcore_ref2ptr(&blkdev->xendev, req->seg[i].gref);
	if (NULL == page)
	    goto err;
	len = (req->seg[i].last_sect - req->seg[i].first_sect + 1) * blkdev->file_blk;
	mem = page + req->seg[i].first_sect * blkdev->file_blk;
	d1printf("%s:   %d: op %d, mem %p, len %zd\n",
		 __FUNCTION__, i, req->operation, mem, len);
	vec[i].iov_base = mem;
	vec[i].iov_len  = len;
	end += len;
    }
    if (end >= blkdev->file_size)
	goto err; /* access beyoind end of file */

    /* run i/o */
    lseek(blkdev->file, start, SEEK_SET);
    switch (req->operation) {
    case BLKIF_OP_READ:
	len = readv(blkdev->file, vec, req->nr_segments);
	break;
    case BLKIF_OP_WRITE_BARRIER:
	fdatasync(blkdev->file);
	len = writev(blkdev->file, vec, req->nr_segments);
	fdatasync(blkdev->file);
	break;
    case BLKIF_OP_WRITE:
	len = writev(blkdev->file, vec, req->nr_segments);
	break;
    default:
	goto err; /* unknown operation */
    }
    if (len != end - start)
	goto err; /* I/O error */

    blk_send_response(blkdev, req->id, req->operation, BLKIF_RSP_OKAY);
    return;

err:
    blk_send_response(blkdev, req->id, req->operation, BLKIF_RSP_ERROR);
    return;
}

static void blk_check_ring(struct blkdev *blkdev)
{
    RING_IDX rc, rp;
    blkif_request_t req;

    if (blkdev->state != BLKIF_CONNECTED)
	return;

    for (;;) {
	blkdev->more_work = 0;
	rc = blkdev->rings.common.req_cons;
	rp = blkdev->rings.common.sring->req_prod;
	rmb(); /* Ensure we see queued requests up to 'rp'. */
	
	while ((rc != rp)) {
	    if (RING_REQUEST_CONS_OVERFLOW(&blkdev->rings.common, rc))
		break;
	    switch (blkdev->protocol) {
	    case BLKIF_PROTOCOL_NATIVE:
		memcpy(&req, RING_GET_REQUEST(&blkdev->rings.native, rc), sizeof(req));
		break;
	    case BLKIF_PROTOCOL_X86_32:
		blkif_get_x86_32_req(&req, RING_GET_REQUEST(&blkdev->rings.x86_32, rc));
		break;
	    case BLKIF_PROTOCOL_X86_64:
		blkif_get_x86_64_req(&req, RING_GET_REQUEST(&blkdev->rings.x86_64, rc));
		break;
	    }
	    blkdev->rings.common.req_cons = ++rc;
	    blk_handle_request(blkdev, &req);
	}
	if (!blkdev->more_work)
	    break;
    }
}

static int blk_alloc(struct xendev *xendev)
{
    struct blkdev *blkdev = container_of(xendev, struct blkdev, xendev);

    d1printf("%s: %p\n", __FUNCTION__, blkdev);
    blkdev->state = BLKIF_PROBED;
    return 0;
}

static int blk_setup(struct xendev *xendev)
{
    struct blkdev *blkdev = container_of(xendev, struct blkdev, xendev);

    d1printf("%s: %p\n", __FUNCTION__, blkdev);
    blk_setup_backend(blkdev);
    blk_connect_frontend(blkdev);
    blk_check_ring(blkdev);
    return 0;
}

static int blk_backend(struct xendev *xendev, char *node, char *val)
{
    struct blkdev *blkdev = container_of(xendev, struct blkdev, xendev);

    d2printf("%s: %p \"%s\" = \"%s\"\n", __FUNCTION__, blkdev, node,
	     val ? val : "<deleted>");
    blk_setup_backend(blkdev);
    return 0;
}

static int blk_frontend(struct xendev *xendev, char *node, char *val)
{
    struct blkdev *blkdev = container_of(xendev, struct blkdev, xendev);

    d2printf("%s: %p \"%s\" = \"%s\"\n", __FUNCTION__, blkdev, node,
	     val ? val : "<deleted>");
    switch (xendev->fe_state) {
    case XenbusStateInitialised:
    case XenbusStateConnected:
	blk_connect_frontend(blkdev);
	blk_check_ring(blkdev);
	break;

    case XenbusStateClosing:
    case XenbusStateClosed:
	blk_disconnect(blkdev, xendev->fe_state);
	break;

    case XenbusStateInitialising:
    case XenbusStateInitWait:
    case XenbusStateUnknown:
	/* make gcc happy */
	break;
    }
    return 0;
}

static int blk_event(struct xendev *xendev)
{
    struct blkdev *blkdev = container_of(xendev, struct blkdev, xendev);

    blk_check_ring(blkdev);
    return 0;
}

static int blk_free(struct xendev *xendev)
{
    struct blkdev *blkdev = container_of(xendev, struct blkdev, xendev);

    d1printf("%s: %p\n", __FUNCTION__, blkdev);

    blk_disconnect(blkdev, 0);
    if (blkdev->params)
	free(blkdev->params);
    if (blkdev->mode)
	free(blkdev->mode);
    return 0;
}

static struct devops blkdev_ops = {
    .size  = sizeof(struct blkdev),
    .alloc = blk_alloc,
    .setup = blk_setup,
    .xs_be = blk_backend,
    .xs_fe = blk_frontend,
    .event = blk_event,
    .free  = blk_free,
};

/* ------------------------------------------------------------- */

static void usage(FILE *fp)
{
    fprintf(fp,
	    "\n"
	    "blkbackd -- no usage yet, sorry\n"
	    "\n");
}

int main(int argc, char *argv[])
{
    struct sigaction act,old;
    int maxfd, fd;
    fd_set set;
    int c;
                
    struct xs_handle *xs = NULL;

    for (;;) {
        if (-1 == (c = getopt(argc, argv, "hdp:")))
            break;
        switch (c) {
        case 'd':
	    debug++;
	    break;
	case 'p':
	    pidfile = optarg;
	    break;

        case 'h':
            usage(stdout);
            exit(0);
        default:
            usage(stderr);
            exit(1);
        }
    }

    /* setup signal handler */
    memset(&act,0,sizeof(act));
    sigemptyset(&act.sa_mask);
    act.sa_handler = catchsig;
    sigaction(SIGTERM,&act,&old);
    sigaction(SIGINT,&act,&old);

    evtchnd = xenner_evtchn_open();
    if (evtchnd < 0) {
	fprintf(stderr, "can't connect to evtchnd\n");
	exit(1);
    }

    xs = xs_daemon_open();
    if (!xs) {
	fprintf(stderr, "can't connect to xenstored\n");
	exit(1);
    }

    /* fork into background, handle pidfile */
    daemonize();

    xenstore_scan(xs, BE_ROOT, &blkdev_ops);

    /* main loop */
    for (;;) {
	if (termsig)
	    break;
	
	FD_ZERO(&set);
	maxfd = 0;

	fd = xs_fileno(xs);
	FD_SET(fd, &set);
	if (maxfd < fd)
	    maxfd = fd;

	fd = xenner_evtchn_fd(evtchnd);
	FD_SET(fd, &set);
	if (maxfd < fd)
	    maxfd = fd;

	switch (select(maxfd+1, &set, NULL, NULL, NULL)) {
	case -1:
	    if (EINTR == errno)
		continue; /* termsig check */
	    perror("select");
	    break;
	case 0:
	    break;
	default:
	    break;
	}

	if (FD_ISSET(xs_fileno(xs), &set))
	    xenstore_update(xs);
	if (FD_ISSET(xenner_evtchn_fd(evtchnd), &set))
	    event_received();
    }
    return 0;
}
