#include "emu32.h"

/* --------------------------------------------------------------------- */
/* inline asm bits                                                       */

int is_pae(void)
{
    static int pae = -1;

    if (-1 == pae) {
	pae = (read_cr4() & X86_CR4_PAE);
	printk("pae is turned %s\n", pae ? "on" : "off");
    }
    return pae;
}

static inline void vmexit_panic(void)
{
    asm volatile("outb %%al,$0xeb"
		 : : : "memory");
}

/* --------------------------------------------------------------------- */
/* helper functions                                                      */

int panic(char *message)
{
    printk("panic: %s\n", message);
    vmexit_panic();
    return 0;
}

static int bounce_trap(struct regs_32 *regs, int trapno, int cbno)
{
    uint32_t *kesp, eip, cs;
    int stack_switch = 0;
    int error_code = 0;
    int interrupt = 0;
    int k = 0;

    if (trapno >= 0) {
	/* trap bounce */
	eip  = xen_ctxt.trap_ctxt[trapno].address;
	cs   = xen_ctxt.trap_ctxt[trapno].cs;
	if (TI_GET_IF(&xen_ctxt.trap_ctxt[trapno]))
	    interrupt = 1;
	switch (trapno) {
	case 14:
	    /* page fault */
	    shared_info.vcpu_info[0].arch.cr2 = read_cr2();
	    error_code = 1;
	    break;
	}
    }
    if (cbno >= 0) {
	/* callback */
	eip  = xencb[cbno].eip;
	cs   = xencb[cbno].cs;
	switch (cbno) {
	case CALLBACKTYPE_event:
 	    interrupt = 1;
	    break;
	}
    }

    if (!cs)
	panic("no guest trap handler");

    if ((regs->cs & 0x03) < (cs & 0x03))
	panic("bounce trap: illegal ring switch\n");
    if ((regs->cs & 0x03) > (cs & 0x03))
	stack_switch = 1;

    /* prepare guest stack: copy from emu, so the handler
     * jumps straigt back without round-trip via emu */
    if (stack_switch) {
	kesp = (void*)xen_tss.esp1;
	kesp[-(++k)] = regs->ss;    // push ss
	kesp[-(++k)] = regs->esp;   // push esp
    } else {
	kesp = (void*)regs->esp;
    }

    kesp[-(++k)] = regs->eflags;    // push eflags
    kesp[-(++k)] = regs->cs;        // push cs
    kesp[-(++k)] = regs->eip;       // push eip
    if (error_code)
	kesp[-(++k)] = regs->error; // push error code

    /* prepare emu stack, so iret jumps to the kernels handler. */
    regs->eip     = eip;
    regs->cs      = cs;
    regs->eflags &= EFLAGS_TRAPMASK;
    if (stack_switch) {
	regs->ss  = xen_tss.ss1;
	regs->esp = xen_tss.esp1;
    }
    regs->esp -= 4*k;

    return 0;
}

static void print_gpf_info(struct regs_32 *regs)
{
    printk("%s: index 0x%x%s%s%s, rip %x\n", __FUNCTION__,
	   (int)(regs->error >> 3),
	   (regs->error & 0x04) ? ", TI"  : "",
	   (regs->error & 0x02) ? ", IDT" : "",
	   (regs->error & 0x01) ? ", EXT" : "",
	   regs->eip);
}

static void print_page_fault_info(struct regs_32 *regs)
{
    printk("%s:%s%s%s%s%s, cs:eip %x:%x, cr2 %x\n", __FUNCTION__,
	   regs->error & 0x01 ? " preset"  : " nopage",
	   regs->error & 0x02 ? " write"   : " read",
	   regs->error & 0x04 ? " user"    : " kernel",
	   regs->error & 0x08 ? " reserved-bit"  : "",
	   regs->error & 0x10 ? " instr-fetch"   : "",
	   regs->cs, regs->eip, read_cr2());
}

/* --------------------------------------------------------------------- */
/* called from assembler                                                 */

asmlinkage void do_boot(struct regs_32 regs)
{
    /* register setup */
    regs.eax    = xen_ctxt.user_regs.eax;
    regs.ebx    = xen_ctxt.user_regs.ebx;
    regs.ecx    = xen_ctxt.user_regs.ecx;
    regs.edx    = xen_ctxt.user_regs.edx;
    regs.esi    = xen_ctxt.user_regs.esi;
    regs.edi    = xen_ctxt.user_regs.edi;
    regs.ebp    = xen_ctxt.user_regs.ebp;
    regs.eip    = xen_ctxt.user_regs.eip;
    regs.cs     = xen_ctxt.user_regs.cs;
    regs.eflags = xen_ctxt.user_regs.eflags;
    regs.esp    = xen_ctxt.user_regs.esp;
    regs.ss     = xen_ctxt.user_regs.ss;

    asm volatile("mov %0, %%ds;\n" :: "r" (xen_ctxt.user_regs.ds) : "memory");
    asm volatile("mov %0, %%es;\n" :: "r" (xen_ctxt.user_regs.es) : "memory");
    asm volatile("mov %0, %%fs;\n" :: "r" (xen_ctxt.user_regs.fs) : "memory");
    asm volatile("mov %0, %%gs;\n" :: "r" (xen_ctxt.user_regs.gs) : "memory");

    /* jump to kernel (via iret in asm code calling us) */
    printk("booting kernel (entry %x:%x) ...\n", regs.cs, regs.eip);
}

asmlinkage void do_illegal_instruction(struct regs_32 regs)
{
    int skip;

    emu_faults[XEN_FAULT_ILLEGAL_INSTRUCTION]++;
    if (context_is_emu(&regs))
	panic("ring0 (emu) illegal instruction");
    if (context_is_user(&regs)) {
	bounce_trap(&regs, 6, -1);
	return;
    }
    
    skip = emulate_32(&regs);
    switch (skip) {
    case -1: /* error */
	panic("instruction emulation failed\n");
	break;
    case 0:  /* bounce to guest */
	bounce_trap(&regs, 6, -1);
	break;
    default: /* handled */
	regs.eip += skip;
	break;
    }
}

asmlinkage void do_general_protection(struct regs_32 regs)
{
    int skip;

    emu_faults[XEN_FAULT_GENERAL_PROTECTION]++;
    if (context_is_emu(&regs)) {
	print_gpf_info(&regs);
	panic("ring0 (emu) general protection fault");
    }
    if (context_is_user(&regs)) {
	bounce_trap(&regs, 13, -1);
	return;
    }
    
    if (0 != regs.error) {
	print_gpf_info(&regs);
	panic("unhandled kernel gpf");
    }

    skip = emulate_32(&regs);
    switch (skip) {
    case -1: /* error */
	panic("instruction emulation failed\n");
	break;
    case 0:  /* bounce to guest */
	bounce_trap(&regs, 13, -1);
	break;
    default: /* handled */
	regs.eip += skip;
	break;
    }
}

asmlinkage void do_page_fault(struct regs_32 regs)
{
    uint32_t cr2 = read_cr2();

    emu_faults[XEN_FAULT_PAGE_FAULT]++;
    if (context_is_emu(&regs)) {
	print_page_fault_info(&regs);
	panic("ring0 (emu) page fault");
    }

    if (regs.error == 3) {
	/* kernel write to r/o page */
	if (!is_pae()) {
	    uint32_t *pte = find_pte_32(cr2);
	    if (*pte & _PAGE_USER)
		goto bounce;
	    /* is kernel page */
	    *pte |= _PAGE_RW;
	} else {
	    uint64_t *pte = find_pte_pae(cr2);
	    if (*pte & _PAGE_USER)
		goto bounce;
	    /* is kernel page */
	    *pte |= _PAGE_RW;
	}
	return;
    }

bounce:
    bounce_trap(&regs, 14, -1);
}

asmlinkage void do_event_callback(struct regs_32 regs)
{
    if (shared_info.vcpu_info[0].evtchn_upcall_mask) {
	printk("%s: IRQs off\n", __FUNCTION__);
	return;
    }
    shared_info.vcpu_info[0].evtchn_upcall_mask = 1;
    bounce_trap(&regs, -1, CALLBACKTYPE_event);
}
