#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <inttypes.h>
#include <errno.h>
#include <assert.h>

#include "xenner.h"
#include "hexdump.h"
#include "mm.h"

/* ------------------------------------------------------------------ */

uint32_t *find_pgd_32(struct xenvm *xen, uint32_t va)
{
    pfn_t mfn;
    int slot;
    uint32_t *pgd;

    mfn  = addr_to_frame(xen->sregs.cr3);
    if (mfn >= xen->pg_total)
	return NULL;
    slot = PGD_INDEX_32(va);
    pgd  = xen->memory + frame_to_addr(mfn);
    d3printf("%s  : va 0x%08" PRIx32 " -> mfn 0x%lx +0x%03x  [%08" PRIx32"]\n",
	     __FUNCTION__, va, mfn, slot, pgd[slot]);
    return pgd + slot;
}

uint32_t *find_pte_32(struct xenvm *xen, uint32_t va)
{
    pfn_t mfn;
    int slot;
    uint32_t *pgd, *pte;

    pgd  = find_pgd_32(xen, va);
    if (!test_pgflag_32(*pgd, _PAGE_PRESENT))
	return NULL;
    mfn  = get_pgframe_32(*pgd);
    if (mfn >= xen->pg_total)
	return NULL;
    slot = PTE_INDEX_32(va);
    pte = xen->memory + frame_to_addr(mfn);
    d3printf("  %s: va 0x%08" PRIx32 " -> mfn 0x%lx +0x%03x  [%08" PRIx32"]\n",
	     __FUNCTION__, va, mfn, slot, pte[slot]);
    return pte + slot;
}

void *find_pa_32(struct xenvm *xen, uint32_t va)
{
    pfn_t mfn;
    int offset;
    uint32_t *pte;

    pte = find_pte_32(xen, va);
    if (NULL == pte)
	return NULL;
    if (!test_pgflag_32(*pte, _PAGE_PRESENT))
	return NULL;
    mfn = get_pgframe_32(*pte);
    if (mfn >= xen->pg_total)
	return NULL;
    offset = addr_offset(va);
    d3printf("   %s: va 0x%08" PRIx32 " -> mfn 0x%lx +0x%03x\n",
	     __FUNCTION__, va, mfn, offset);
    return xen->memory + frame_to_addr(mfn) + offset;
}

int map_region_pse_32(struct xenvm *xen, uint32_t va, uint32_t flags,
		      pfn_t start, pfn_t count)
{
    uint32_t *pgd;
    pfn_t mfn;
    
    d1printf("%s: mfns 0x%lx +0x%lx at 0x%" PRIx32 "\n",
	     __FUNCTION__, start, count, va);
    flags |= _PAGE_PSE;
    pgd = find_pgd_32(xen, va);
    assert(pgd);
    for (mfn = start; mfn < start + count; mfn += PGD_COUNT_32, pgd++)
	*pgd = get_pgentry_32(mfn, flags);
    return 0;
}

int map_linear_pgt_32(struct xenvm *xen, uint32_t va, uint32_t flags)
{
    uint32_t *pgd;
    pfn_t mfn;

    pgd = find_pgd_32(xen, va);
    assert(pgd);
    mfn = addr_to_frame(xen->sregs.cr3);
    d1printf("%s: at 0x%" PRIx32 " mfn 0x%lx\n", __FUNCTION__, va, mfn);
    *pgd = get_pgentry_32(mfn, flags);
    return 0;
}

int map_normal_pgt_32(struct xenvm *xen, uint32_t va, uint32_t flags, pfn_t table)
{
    uint32_t *pgd;
    
    d1printf("%s: mfn 0x%lx at 0x%" PRIx32 "\n",
	     __FUNCTION__, table, va);
    pgd = find_pgd_32(xen, va);
    assert(pgd);
    *pgd = get_pgentry_32(table, flags);
    return 0;
}

/* ------------------------------------------------------------------ */

uint64_t *find_pgd_pae(struct xenvm *xen, uint32_t va)
{
    pfn_t mfn;
    int slot;
    uint64_t *pgd;

    mfn  = addr_to_frame(xen->sregs.cr3);
    if (mfn >= xen->pg_total)
	return NULL;
    slot = PGD_INDEX_PAE(va);
    pgd  = xen->memory + frame_to_addr(mfn);
    d3printf("%s    : va 0x%08" PRIx32 " -> mfn 0x%lx +0x%03x  [%08" PRIx64 "]\n",
	     __FUNCTION__, va, mfn, slot, pgd[slot]);
    return pgd + slot;
}

uint64_t *find_pmd_pae(struct xenvm *xen, uint32_t va)
{
    pfn_t mfn;
    int slot;
    uint64_t *pgd, *pmd;

    pgd = find_pgd_pae(xen, va);
    if (!test_pgflag_pae(*pgd, _PAGE_PRESENT))
	return NULL;
    mfn = get_pgframe_pae(*pgd);
    if (mfn >= xen->pg_total)
	return NULL;
    slot = PMD_INDEX_PAE(va);
    pmd = xen->memory + frame_to_addr(mfn);
    d3printf("  %s  : va 0x%08" PRIx32 " -> mfn 0x%lx +0x%03x  [%08" PRIx64 "]\n",
	     __FUNCTION__, va, mfn, slot, pmd[slot]);
    return pmd + slot;
}

uint64_t *find_pte_pae(struct xenvm *xen, uint32_t va)
{
    pfn_t mfn;
    int slot;
    uint64_t *pmd, *pte;

    pmd = find_pmd_pae(xen, va);
    if (NULL == pmd)
	return NULL;
    if (!test_pgflag_pae(*pmd, _PAGE_PRESENT))
	return NULL;
    mfn = get_pgframe_pae(*pmd);
    if (mfn >= xen->pg_total)
	return NULL;
    slot = PTE_INDEX_PAE(va);
    pte = xen->memory + frame_to_addr(mfn);
    d3printf("    %s: va 0x%08" PRIx32 " -> mfn 0x%lx +0x%03x  [%08" PRIx64 "]\n",
	     __FUNCTION__, va, mfn, slot, pte[slot]);
    return pte + slot;
}

void *find_pa_pae(struct xenvm *xen, uint32_t va)
{
    pfn_t mfn;
    int offset;
    uint64_t *pte;

    pte = find_pte_pae(xen, va);
    if (NULL == pte)
	return NULL;
    if (!test_pgflag_pae(*pte, _PAGE_PRESENT))
	return NULL;
    mfn = get_pgframe_pae(*pte);
    if (mfn >= xen->pg_total)
	return NULL;
    offset = addr_offset(va);
    d3printf("     %s: va 0x%08" PRIx32 " -> mfn 0x%lx +0x%03x\n",
	     __FUNCTION__, va, mfn, offset);
    return xen->memory + frame_to_addr(mfn) + offset;
}

int map_region_pse_pae(struct xenvm *xen, uint32_t va, uint32_t flags,
		       pfn_t start, pfn_t count)
{
    uint64_t *pmd;
    pfn_t mfn;

    d1printf("%s: mfns 0x%lx +0x%lx at 0x%" PRIx32 "\n",
	     __FUNCTION__, start, count, va);
    flags |= _PAGE_PSE;
    pmd = find_pmd_pae(xen, va);
    assert(pmd);
    for (mfn = start; mfn < start + count; mfn += PMD_COUNT_PAE, pmd++)
	*pmd = get_pgentry_pae(mfn, flags);
    return 0;
}

int map_linear_pgt_pae(struct xenvm *xen, uint32_t va, uint32_t flags)
{
    uint64_t *pgd;
    uint64_t *pmd;
    pfn_t mfn;
    int i;

    d1printf("%s: at 0x%" PRIx32 "\n", __FUNCTION__, va);
    pgd = find_pgd_pae(xen, 0);
    pmd = find_pmd_pae(xen, va);
    assert(pmd);

    for (i = 0; i < 4; i++) {
	if (test_pgflag_pae(pgd[i], _PAGE_PRESENT)) {
	    mfn = get_pgframe_pae(pgd[i]);
	    d1printf("%s:    %d mfn 0x%lx\n", __FUNCTION__, i, mfn);
	    pmd[i] = get_pgentry_pae(mfn, flags);
	} else {
	    d1printf("%s:    %d !present\n", __FUNCTION__, i);
	    pmd[i] = 0;
	}
    }
    return 0;
}

int map_normal_pgt_pae(struct xenvm *xen, uint32_t va, uint32_t flags, pfn_t table)
{
    uint64_t *pmd;

    d1printf("%s: mfn 0x%lx at 0x%" PRIx32 "\n",
	     __FUNCTION__, table, va);
    pmd = find_pmd_pae(xen, va);
    assert(pmd);
    *pmd = get_pgentry_pae(table, flags);
    return 0;
}

void dump_pgtables_pae(struct xenvm *xen, uint32_t vaddr)
{
    void *ptr;

    ptr = find_pgd_pae(xen, vaddr);
    hexdump("pgd", ptr, 8);
    ptr = find_pmd_pae(xen, vaddr);
    hexdump("pmd", ptr, 8);
    ptr = find_pte_pae(xen, vaddr);
    hexdump("pte", ptr, 8);
}

/* ------------------------------------------------------------------ */

uint64_t *find_pgd_64(struct xenvm *xen, uint64_t va)
{
    pfn_t mfn;
    int slot;
    uint64_t *pgd;

    mfn  = addr_to_frame(xen->sregs.cr3);
    if (mfn >= xen->pg_total)
	return NULL;
    slot = PGD_INDEX_64(va);
    pgd  = xen->memory + frame_to_addr(mfn);
    d3printf("%s      : va 0x%08" PRIx64 " -> mfn 0x%lx +0x%03x  [%08" PRIx64 "]\n",
	     __FUNCTION__, va, mfn, slot, pgd[slot]);
    return pgd + slot;
}

uint64_t *find_pud_64(struct xenvm *xen, uint64_t va)
{
    pfn_t mfn;
    int slot;
    uint64_t *pgd, *pud;

    pgd = find_pgd_64(xen, va);
    if (!test_pgflag_64(*pgd, _PAGE_PRESENT))
	return NULL;
    mfn = get_pgframe_64(*pgd);
    if (mfn >= xen->pg_total)
	return NULL;
    slot = PUD_INDEX_64(va);
    pud = xen->memory + frame_to_addr(mfn);
    d3printf("  %s    : va 0x%08" PRIx64 " -> mfn 0x%lx +0x%03x  [%08" PRIx64 "]\n",
	     __FUNCTION__, va, mfn, slot, pud[slot]);
    return pud + slot;
}

uint64_t *find_pmd_64(struct xenvm *xen, uint64_t va)
{
    pfn_t mfn;
    int slot;
    uint64_t *pud, *pmd;

    pud = find_pud_64(xen, va);
    if (NULL == pud)
	return NULL;
    if (!test_pgflag_64(*pud, _PAGE_PRESENT))
	return NULL;
    mfn = get_pgframe_64(*pud);
    if (mfn >= xen->pg_total)
	return NULL;
    slot = PMD_INDEX_64(va);
    pmd = xen->memory + frame_to_addr(mfn);
    d3printf("    %s  : va 0x%08" PRIx64 " -> mfn 0x%lx +0x%03x  [%08" PRIx64 "]\n",
	     __FUNCTION__, va, mfn, slot, pmd[slot]);
    return pmd + slot;
}

uint64_t *find_pte_64(struct xenvm *xen, uint64_t va)
{
    pfn_t mfn;
    int slot;
    uint64_t *pmd, *pte;

    pmd = find_pmd_64(xen, va);
    if (NULL == pmd)
	return NULL;
    if (!test_pgflag_64(*pmd, _PAGE_PRESENT))
	return NULL;
    mfn = get_pgframe_64(*pmd);
    if (mfn >= xen->pg_total)
	return NULL;
    slot = PTE_INDEX_64(va);
    pte = xen->memory + frame_to_addr(mfn);
    d3printf("      %s: va 0x%08" PRIx64 " -> mfn 0x%lx +0x%03x  [%08" PRIx64 "]\n",
	     __FUNCTION__, va, mfn, slot, pte[slot]);
    return pte + slot;
}

void *find_pa_64(struct xenvm *xen, uint64_t va)
{
    pfn_t mfn;
    int offset;
    uint64_t *pte;

    pte = find_pte_64(xen, va);
    if (NULL == pte)
	return NULL;
    if (!test_pgflag_64(*pte, _PAGE_PRESENT))
	return NULL;
    mfn = get_pgframe_64(*pte);
    if (mfn >= xen->pg_total)
	return NULL;
    offset = addr_offset(va);
    d3printf("       %s: va 0x%08" PRIx64 " -> mfn 0x%lx +0x%03x\n",
	     __FUNCTION__, va, mfn, offset);
    return xen->memory + frame_to_addr(mfn) + offset;
}

static pfn_t alloc_pgtable(struct xenvm *xen)
{
    pfn_t mfn;
    
    if (PGTABLES_64 == xen->idx_pgtables) {
	d0printf("%s: out of page table pages (%d)\n",
		 __FUNCTION__, xen->idx_pgtables);
	exit(1);
    }
    mfn = xen->mfn_pgtables + xen->idx_pgtables++;
    d1printf("%s: mfn 0x%lx (%d/%d)\n", __FUNCTION__,
	     mfn, xen->idx_pgtables, PGTABLES_64);
    return mfn;
}

static uint64_t *alloc_pmd_64(struct xenvm *xen, uint64_t va)
{
    uint64_t *pud;
    pfn_t mfn;

    pud = find_pud_64(xen, va);
    if (NULL == pud) {
	uint64_t *pgd = find_pgd_64(xen, va);
	mfn = alloc_pgtable(xen);
	*pgd = get_pgentry_64(mfn, PGT_PGFLAGS_64);
	pud = find_pud_64(xen, va);
    }
    mfn = alloc_pgtable(xen);
    *pud = get_pgentry_64(mfn, PGT_PGFLAGS_64);
    return find_pmd_64(xen, va);
}

int map_region_pse_64(struct xenvm *xen, uint64_t va, uint32_t flags,
		      pfn_t start, pfn_t count)
{
    uint64_t *pmd;
    pfn_t mfn;

    d1printf("%s: mfns 0x%lx +0x%lx at 0x%" PRIx64 "\n",
	     __FUNCTION__, start, count, va);
    flags |= _PAGE_PSE;
    for (mfn = start; mfn < start + count; mfn += PMD_COUNT_64, pmd++) {
	pmd = find_pmd_64(xen, va + frame_to_addr(mfn-start));
	if (NULL == pmd)
	    pmd = alloc_pmd_64(xen, va + frame_to_addr(mfn-start));
	*pmd = get_pgentry_64(mfn, flags);
    }
    return 0;
}

int map_linear_pgt_64(struct xenvm *xen, uint64_t va, uint32_t flags)
{
    uint64_t *pgd;
    pfn_t mfn;

    pgd = find_pgd_64(xen, va);
    assert(pgd);
    mfn = addr_to_frame(xen->sregs.cr3);
    d1printf("%s: at 0x%" PRIx64 " mfn 0x%lx\n", __FUNCTION__, va, mfn);
    *pgd = get_pgentry_64(mfn, flags);
    return 0;
}

/* ------------------------------------------------------------------ */

void *mfn_to_ptr(struct xenvm *xen, uint64_t mfn)
{
    if (mfn >= xen->pg_total)
	return NULL;
    return xen->memory + frame_to_addr(mfn);
}

void *guest_paddr_to_ptr(struct xenvm *xen, uint64_t addr)
{
    uint64_t mfn = xen->mfn_guest + addr_to_frame(addr);
    void *ptr = mfn_to_ptr(xen, mfn);
    return ptr ? ptr + addr_offset(addr) : NULL;
}

void *guest_vaddr_to_ptr(struct xenvm *xen, uint64_t addr)
{
    switch (xen->mode) {
    case XENMODE_32:
	return find_pa_32(xen, addr);
    case XENMODE_PAE:
	return find_pa_pae(xen, addr);
    default:
	return find_pa_64(xen, addr);
    }
}

void *emu_paddr_to_ptr(struct xenvm *xen, uint64_t addr)
{
    return mfn_to_ptr(xen, xen->mfn_emu) + addr;
}

void *emu_vaddr_to_ptr(struct xenvm *xen, uint64_t addr)
{
    return emu_paddr_to_ptr(xen, addr - xen->emu_va);
}

/* ------------------------------------------------------------------ */

int copy_from_guest(struct xenvm *xen, void *dest, uint64_t vaddr, size_t size)
{
    int offset;
    int bytes;
    void *src;
    
    while (size > 0) {
	offset = addr_offset(vaddr);
	bytes  = PAGE_SIZE - offset;
	if (bytes > size)
	    bytes = size;
	src = guest_vaddr_to_ptr(xen, vaddr);
	if (NULL == src)
	    return -EFAULT;
	memcpy(dest, src, bytes);
	dest  += bytes;
	vaddr += bytes;
	size  -= bytes;
    }
    return 0;
}

int copy_to_guest(struct xenvm *xen, uint64_t vaddr, void *src, size_t size)
{
    int offset;
    int bytes;
    void *dest;
    
    while (size > 0) {
	offset = addr_offset(vaddr);
	bytes  = PAGE_SIZE - offset;
	if (bytes > size)
	    bytes = size;
	dest = guest_vaddr_to_ptr(xen, vaddr);
	if (NULL == dest)
	    return -EFAULT;
	memcpy(dest, src, bytes);
	src   += bytes;
	vaddr += bytes;
	size  -= bytes;
    }
    return 0;
}
