#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <inttypes.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include <xs.h>
#include "xennerctrl.h"
#include <xen/grant_table.h>
#include <xen/io/xenbus.h>

#include "xenbackd.h"
#include "daemon.h"
#include "shared.h"

/* ------------------------------------------------------------- */

static LIST_HEAD(xendevs);
static int devcount;

int evtchnd;

/* ------------------------------------------------------------- */

static struct xendev *find_xendev(int dom, int dev)
{
    struct xendev *xendev;
    struct list_head *item;

    list_for_each(item, &xendevs) {
	xendev = list_entry(item, struct xendev, next);
	if (xendev->dom == dom  &&  xendev->dev == dev)
	    return xendev;
    }
    return NULL;
}

static struct xendev *get_xendev(int dom, int dev,
				 char *be_root,
				 struct xs_handle *xs,
				 struct devops *ops)
{
    struct xendev *xendev;

    xendev = find_xendev(dom, dev);
    if (!xendev) {
	xendev = malloc(ops->size);
	memset(xendev,0,ops->size);
	xendev->dom = dom;
	xendev->dev = dev;

	xendev->be_root = be_root;
	xendev->xs      = xs;
	xendev->ops     = ops;

	list_add_tail(&xendev->next, &xendevs);
	devcount++;
	d1printf("%s: %d/%d\n", __FUNCTION__,
		 xendev->dom, xendev->dev);

	if (xendev->ops->alloc)
	    xendev->ops->alloc(xendev);
    }
    return xendev;
}

static struct xendev *del_xendev(int dom, int dev)
{
    struct xendev *xendev;
    struct list_head *item, *tmp;

    list_for_each_safe(item, tmp, &xendevs) {
	xendev = list_entry(item, struct xendev, next);
	if (xendev->dom != dom)
	    continue;
	if (xendev->dev != dev && dev != -1)
	    continue;

	if (xendev->ops->free)
	    xendev->ops->free(xendev);

	if (xendev->fe) {
	    char token[BUFSIZE];
	    snprintf(token, sizeof(token), "fe:%p", xendev);
	    xs_unwatch(xendev->xs, xendev->fe, token);
	    free(xendev->fe);
	}
	
	d1printf("%s: %d/%d\n", __FUNCTION__,
		 xendev->dom, xendev->dev);
	list_del(&xendev->next);
	devcount--;
	free(xendev);
    }
    return NULL;
}
/* ------------------------------------------------------------- */

int write_be_str(struct xendev *xendev, char *node, char *val)
{
    char abspath[BUFSIZE];
    
    snprintf(abspath, sizeof(abspath), "%s/%d/%d/%s",
	     xendev->be_root, xendev->dom, xendev->dev, node);
    return xs_write(xendev->xs, 0, abspath, val, strlen(val));
}

int write_be_int(struct xendev *xendev, char *node, int ival)
{
    char val[32];

    snprintf(val, sizeof(val), "%d", ival);
    return write_be_str(xendev, node, val);
}

char *read_be_str(struct xendev *xendev, char *node)
{
    char abspath[BUFSIZE];
    unsigned int len;

    snprintf(abspath, sizeof(abspath), "%s/%d/%d/%s",
	     xendev->be_root, xendev->dom, xendev->dev, node);
    return xs_read(xendev->xs, 0, abspath, &len);
}

int read_be_int(struct xendev *xendev, char *node)
{
    char *val;
    int ival;

    val = read_be_str(xendev, node);
    ival = val ? atoi(val) : 0;
    free(val);
    return ival;
}

char *read_fe_str(struct xendev *xendev, char *node)
{
    char abspath[BUFSIZE];
    unsigned int len;

    snprintf(abspath, sizeof(abspath), "%s/%s",
	     xendev->fe, node);
    return xs_read(xendev->xs, 0, abspath, &len);
}

int read_fe_int(struct xendev *xendev, char *node)
{
    char *val;
    int ival;

    val = read_fe_str(xendev, node);
    ival = val ? atoi(val) : 0;
    free(val);
    return ival;
}

static int check_state_xendev(struct xendev *xendev, int be, int fe)
{
    enum xenbus_state state;

    if (be) {
	state = read_be_int(xendev, "state");
	if (xendev->be_state != state) {
	    d1printf("%s: be: %d -> %d\n", __FUNCTION__,
		     xendev->be_state, state);
	    xendev->be_state  = state;
	}
    }

    if (fe) {
	state = read_fe_int(xendev, "state");
	if (xendev->fe_state != state) {
	    d1printf("%s: fe: %d -> %d\n", __FUNCTION__,
		     xendev->fe_state, state);
	    xendev->fe_state  = state;
	}
    }
    return 0;
}

int change_state_xendev(struct xendev *xendev, enum xenbus_state state)
{
    write_be_int(xendev, "state", state);
    d1printf("%s: be: %d -> %d\n", __FUNCTION__,
	     xendev->be_state, state);
    xendev->be_state  = state;
    return 0;
}

static int setup_xendev(struct xendev *xendev)
{
    char token[BUFSIZE];

    if (NULL != xendev->fe)
	/* already done */
	return 0;

    xendev->fe = read_be_str(xendev, "frontend");
    if (NULL == xendev->fe)
	/* no fe ptr yet */
	return 0;

    /* setup frontend watch */
    snprintf(token, sizeof(token), "fe:%p", xendev);
    xs_watch(xendev->xs, xendev->fe, token);

    check_state_xendev(xendev, 1, 1);

    if (xendev->ops->setup)
	xendev->ops->setup(xendev);
    return 0;
}

/* ------------------------------------------------------------- */

int vmcore_map(struct xendev *xendev)
{
    char filename[256];
    struct stat st;
    char *name;
    int i, fd;

    snprintf(filename, sizeof(filename), "/var/run/xenner/vmcore.%d",
	     xendev->dom);
    fd = open(filename, O_RDWR);
    if (-1 == fd)
	return -1;
    fstat(fd, &st);
    xendev->vmcore_size = st.st_size;
    xendev->vmcore = mmap(NULL, xendev->vmcore_size,
			  PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
    if ((void*)-1 == xendev->vmcore) {
	d1printf("%s: mmap(%s,%zd): %s\n",
		 __FUNCTION__, filename, xendev->vmcore_size, strerror(errno));
	xendev->vmcore = NULL;
	close(fd);
	return -1;
    }
    close(fd);

    xendev->vm_ehdr  = xendev->vmcore;
    xendev->vm_phdrs = xendev->vmcore + xendev->vm_ehdr->e_phoff;
    xendev->vm_shdrs = xendev->vmcore + xendev->vm_ehdr->e_shoff;
    xendev->vm_shstrtab = xendev->vmcore +
	xendev->vm_shdrs[xendev->vm_ehdr->e_shstrndx].sh_offset;

    for (i = 0; i < xendev->vm_ehdr->e_shnum; i++) {
	name = xendev->vm_shstrtab + xendev->vm_shdrs[i].sh_name;
	d1printf("  %d: %s\n", i, name);
	if (0 == strcmp(name, ".addr.grant_table"))
	    xendev->grant_table = xendev->vmcore + xendev->vm_shdrs[i].sh_offset;
    }
	
    return 0;
}

int vmcore_unmap(struct xendev *xendev)
{
    munmap(xendev->vmcore, xendev->vmcore_size);
    xendev->vmcore      = NULL;
    xendev->grant_table = NULL;
    return 0;
}

void *vmcore_mfn2ptr(struct xendev *xendev, uint64_t mfn)
{
    uint64_t offset;
    void *ptr;

    if (!xendev->vmcore)
	return NULL;
    offset = VMCORE_HEADER + mfn * PAGE_SIZE;
    if (offset >= xendev->vmcore_size)
	return NULL;
    ptr = xendev->vmcore + offset;
    d2printf("%s: mfn 0x%" PRIx64 " -> %p\n", __FUNCTION__, mfn, ptr);
    return ptr;
}

void *vmcore_ref2ptr(struct xendev *xendev, uint32_t ref)
{
    struct grant_entry *e;
    
    if (!xendev->grant_table)
	return NULL;

    e = xendev->grant_table + ref;
    if ((e->flags & GTF_type_mask) != GTF_permit_access)
	goto err;
    
    return vmcore_mfn2ptr(xendev, e->frame);

err:
    d1printf("%s: flags 0x%x, domid %d, frame 0x%x FAILED\n", __FUNCTION__,
	     (int)xendev->grant_table[ref].flags,
	     (int)xendev->grant_table[ref].domid,
	     (int)xendev->grant_table[ref].frame);
    return NULL;
}

/* ------------------------------------------------------------- */

void xenstore_scan(struct xs_handle *xs, char *be_root,
		   struct devops *ops)
{
    struct xendev *xendev;
    char path[BUFSIZE], token[BUFSIZE];
    char **dom = NULL;
    char **dev = NULL;
    unsigned int cdom = 0, cdev = 0, i, j;
        
    /* setup watch */
    snprintf(token, sizeof(token), "be:%p:%p", be_root, ops);
    xs_watch(xs, be_root, token);

    /* look for backends */
    d1printf("%s: %s\n", __FUNCTION__, be_root);
    dom = xs_directory(xs, 0, be_root, &cdom);
    for (i = 0; i < cdom; i++) {
	d1printf("%s:   %s\n", __FUNCTION__, dom[i]);
	snprintf(path, sizeof(path), "%s/%s", be_root, dom[i]);
	dev = xs_directory(xs, 0, path, &cdev);
	for (j = 0; j < cdev; j++) {
	    d1printf("%s:     %s\n", __FUNCTION__, dev[j]);
	    xendev = get_xendev(atoi(dom[i]), atoi(dev[j]), be_root, xs, ops);
	    setup_xendev(xendev);
	}
	if (dev)
	    free(dev);
    }
    if (dom)
	free(dom);
}

static void xenstore_update_be(struct xs_handle *xs, char *watch,
			       char *be_root, struct devops *ops)
{
    struct xendev *xendev;
    char path[BUFSIZE], *val;
    int dom, dev;
    unsigned int len;

    len = strlen(be_root);
    if (0 != strncmp(be_root, watch, strlen(be_root)))
	return;
    if (3 != sscanf(watch+len, "/%d/%d/%255s", &dom, &dev, path)) {
	strcpy(path, "");
	if (2 != sscanf(watch+len, "/%d/%d", &dom, &dev)) {
	    dev = -1;
	    if (1 != sscanf(watch+len, "/%d", &dom))
		return;
	}
    }
    val = xs_read(xs, 0, watch, &len);

    if (NULL == val) {
	del_xendev(dom, dev);
	return;
    }

    if (-1 != dev) {
	xendev = get_xendev(dom, dev, be_root, xs, ops);
	setup_xendev(xendev);
	if (xendev->fe && xendev->ops->xs_be)
	    xendev->ops->xs_be(xendev, path, val);
    }

    if (val)
	free(val);
}

static void xenstore_update_fe(struct xs_handle *xs, char *watch, struct xendev *xendev)
{
    char *val, *node;
    unsigned int len;

    val = xs_read(xs, 0, watch, &len);

    node = strrchr(watch, '/');
    if (node) {
	node++;
	if (0 == strcmp(node, "state"))
	    check_state_xendev(xendev, 0, 1);
    }
    if (xendev->ops->xs_fe)
	xendev->ops->xs_fe(xendev, node, val);

    if (val)
	free(val);
}

void xenstore_update(struct xs_handle *xs)
{
    char **vec = NULL;
    intptr_t ptr, be;
    unsigned int count;
        
    vec = xs_read_watch(xs, &count);
    if (NULL == vec)
	goto cleanup;
    d2printf("%s: %s [%s]\n", __FUNCTION__,
	     vec[XS_WATCH_PATH], vec[XS_WATCH_TOKEN]);

    if (2 == sscanf(vec[XS_WATCH_TOKEN], "be:%" PRIxPTR ":%" PRIxPTR, &be, &ptr))
	xenstore_update_be(xs, vec[XS_WATCH_PATH], (void*)be, (void*)ptr);
    if (1 == sscanf(vec[XS_WATCH_TOKEN], "fe:%" PRIxPTR, &ptr))
	xenstore_update_fe(xs, vec[XS_WATCH_PATH], (void*)ptr);

cleanup:
    if (vec)
	free(vec);
}

void event_received(void)
{
    struct xendev *xendev;
    struct list_head *item;
    int port;
    
    port = xenner_evtchn_pending(evtchnd);
    if (-1 == port) {
	d1printf("%s: evtchnd failure, huh?\n", __FUNCTION__);
	exit(1);
    }

    list_for_each(item, &xendevs) {
	xendev = list_entry(item, struct xendev, next);
	if (xendev->local_port != port)
	    continue;
	d2printf("%s: port %d -> dom %d, dev %d\n", __FUNCTION__,
		 port, xendev->dom, xendev->dev);
	if (xendev->ops->event)
	    xendev->ops->event(xendev);
	return;
    }
    d1printf("%s: port %d -> Huh, not found?\n", __FUNCTION__, port);
}
