#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <inttypes.h>
#include <errno.h>

#include <xs.h>
#include <xen/io/xenbus.h>
#include <xen/io/xs_wire.h>

#include "xenner.h"
#include "mm.h"

/* ------------------------------------------------------------------ */
/* fallback implementation -- just return EIO                         */

static int xs_send_message(struct xenvm *xen, uint8_t *msg, int len)
{
    uint64_t paddr = frame_to_addr(xen->xenstore_pfn);
    struct xenstore_domain_interface *xs = guest_paddr_to_ptr(xen, paddr);
    XENSTORE_RING_IDX cons, prod, i;

    cons = xs->rsp_cons;
    prod = xs->rsp_prod;

    if (prod - cons > sizeof(xs->rsp))
	return -EIO;

    if (sizeof(xs->rsp) - (prod - cons) -1 < len)
	return -EAGAIN;

    /* FIXME: mb */

    for (i = 0; i < len; i++, prod++)
	xs->rsp[MASK_XENSTORE_IDX(prod)] = msg[i];

    /* FIXME: wmb */

    xs->rsp_prod += len;

    raise_event(xen, 0, xen->xenstore_event);
    return 0;
}

static void xs_print_message(struct xenvm *xen, char *prefix, uint8_t *msg, int len)
{
    struct xsd_sockmsg *hdr = (void*)msg;

    switch (hdr->type) {
    case XS_DIRECTORY:
	d0printf("%s: directory \"%s\"\n", prefix, msg+16);
	break;
    case XS_READ:
	d0printf("%s: read \"%s\"\n", prefix, msg+16);
	break;
    case XS_WATCH:
	d0printf("%s: watch \"%s\"\n", prefix, msg+16);
	break;
    case XS_TRANSACTION_START:
	d0printf("%s: transaction end\n", prefix);
	break;
    case XS_TRANSACTION_END:
	d0printf("%s: transaction end\n", prefix);
	break;
    case XS_ERROR:
	d0printf("%s: error \"%s\"\n", prefix, msg+16);
	break;
    default:
	d0printf("%s: unknown type %d\n", prefix, hdr->type);
    }
#if 0
    hexdump(prefix, msg, len);
#endif
}

static int xs_handle_message(struct xenvm *xen, uint8_t *msg, int len)
{
    uint8_t reply[32];
    struct xsd_sockmsg *hdr = (void*)msg;
    struct xsd_sockmsg *re  = (void*)reply;

    xs_print_message(xen, "xenbus << ", msg, len);

    memset(&reply, 0, sizeof(reply));
    re->type   = XS_ERROR;
    re->req_id = hdr->req_id;
    re->len    = 4;
    memcpy(reply + 16, "EIO", re->len);

    xs_print_message(xen, "xenbus >> ", reply, re->len + 16);
    xs_send_message(xen, reply, re->len + 16);
    return 0;
}

static int xenstore_event_internal(struct xenvm *xen)
{
    uint64_t paddr = frame_to_addr(xen->xenstore_pfn);
    struct xenstore_domain_interface *xs = guest_paddr_to_ptr(xen, paddr);
    XENSTORE_RING_IDX cons, prod, i;
    uint8_t buf[sizeof(xs->req)+1];
    struct xsd_sockmsg *hdr = (void*)buf;
    
    cons = xs->req_cons;
    prod = xs->req_prod;
    /* FIXME: mb */

    if (cons == prod || (prod - cons) > sizeof(xs->req))
	return 0;

    for (;;) {
	if (prod - cons < 16) {
	    d2printf("%s: msg head incomplete (%d/16)\n", __FUNCTION__,
		     prod - cons);
	    break;
	}
	for (i = 0; i < 16; i++, cons++)
	    buf[i] = xs->req[MASK_XENSTORE_IDX(cons)];
	hdr = (void*)buf;
	if (prod - cons < hdr->len) {
	    d2printf("%s: msg body incomplete (%d/%d)\n", __FUNCTION__,
		     prod - cons, hdr->len);
	    break;
	}
	for (; i < hdr->len + 16; i++, cons++)
	    buf[i] = xs->req[MASK_XENSTORE_IDX(cons)];
	buf[i] = 0;

	/* FIXME: mb */
	xs->req_cons = cons;

	xs_handle_message(xen, buf, i);
    }
    return 0;
}

/* ------------------------------------------------------------------ */

#define UUID0 "00000000-0000-0000-0000-000000000000"

static int xenstore_setup_host(struct xenvm *xen)
{
    char name[]  = "xenner host";
    char domid[] = "0";
    char vm0[]   = "/vm/" UUID0;
    char uuid0[] = UUID0;
    unsigned int len;
    char *val;

    val = xs_read(xen->xenstore, 0, "/local/domain/0/name", &len);
    if (NULL != val) {
	/* already done */
	free(val);
	return 0;
    }

    xs_write(xen->xenstore, 0, "/local/domain/0/name",  name,  sizeof(name));
    xs_write(xen->xenstore, 0, "/local/domain/0/domid", domid, sizeof(domid));
    xs_write(xen->xenstore, 0, "/local/domain/0/vm",    vm0,   sizeof(vm0));

    xs_write(xen->xenstore, 0, "/vm/" UUID0 "/name",    name,  sizeof(name));
    xs_write(xen->xenstore, 0, "/vm/" UUID0 "/uuid",    uuid0, sizeof(uuid0));
    return 0;
}

static int xenstore_domain_set_str(struct xenvm *xen, char *relpath, char *val)
{
    char abspath[256];

    snprintf(abspath, sizeof(abspath), "/local/domain/%d/%s", xen->domid, relpath);
    if (!xs_write(xen->xenstore, 0, abspath, val, strlen(val)))
	d0printf("xs_write %s: failed\n", abspath);
    return 0;
}

static int xenstore_domain_set_int(struct xenvm *xen, char *relpath, int64_t ival)
{
    char val[32];

    snprintf(val, sizeof(val), "%" PRId64, ival);
    return xenstore_domain_set_str(xen, relpath, val);
}

static int xenstore_vm_set_str(struct xenvm *xen, char *relpath, char *val)
{
    char abspath[256];

    snprintf(abspath, sizeof(abspath), "%s/%s", xen->vm, relpath);
    if (!xs_write(xen->xenstore, 0, abspath, val, strlen(val)))
	d0printf("xs_write %s: failed\n", abspath);
    return 0;
}

static int xenstore_vm_set_int(struct xenvm *xen, char *relpath, int64_t ival)
{
    char val[32];

    snprintf(val, sizeof(val), "%" PRId64, ival);
    return xenstore_vm_set_str(xen, relpath, val);
}

static int xenstore_domain_permit(struct xenvm *xen)
{
    struct xs_permissions perms = {
	.id    = xen->domid,
	.perms = XS_PERM_WRITE,
    };
    char abspath[256];

    snprintf(abspath, sizeof(abspath), "/local/domain/%d", xen->domid);

    if (!xs_mkdir(xen->xenstore, 0, abspath)) {
	d0printf("%s: xs_mkdir %s: failed\n", __FUNCTION__, abspath);
	return -1;
    }
    if (!xs_set_permissions(xen->xenstore, 0, abspath, &perms, 1)) {
	d0printf("%s: xs_set_permissions failed\n", __FUNCTION__);
	return -1;
    }
    return 0;
}

static char *xenstore_device_protocol(struct xenvm *xen)
{
    static char *x86_32 = "x86_32-abi";
    static char *x86_64 = "x86_64-abi";
    
    switch (xen->mode) {
    case XENMODE_32:
    case XENMODE_PAE:
	return x86_32;
    case XENMODE_64:
	return x86_64;
    default:
	return NULL;
    }
}

static int xenstore_device_set_str(struct xenvm *xen, char *dev, char *node, char *val)
{
    char abspath[256];

    snprintf(abspath, sizeof(abspath), "%s/%s", dev, node);
    if (!xs_write(xen->xenstore, 0, abspath, val, strlen(val)))
	d0printf("xs_write %s: failed\n", abspath);
    return 0;
}

static int xenstore_device_set_int(struct xenvm *xen, char *dev, char *node, int ival)
{
    char abspath[256];
    char val[32];

    snprintf(abspath, sizeof(abspath), "%s/%s", dev, node);
    snprintf(val, sizeof(val), "%d", ival);
    if (!xs_write(xen->xenstore, 0, abspath, val, strlen(val)))
	d0printf("xs_write %s: failed\n", abspath);
    return 0;
}

static int xenstore_device_mkdir(struct xenvm *xen, char *dev, int p)
{
    struct xs_permissions perms = {
	.id    = xen->domid,
	.perms = p,
    };

    if (!xs_mkdir(xen->xenstore, 0, dev)) {
	d0printf("xs_mkdir %s: failed\n", dev);
	return -1;
    }

    if (!xs_set_permissions(xen->xenstore, 0, dev, &perms, 1)) {
	d0printf("%s: xs_set_permissions failed\n", __FUNCTION__);
	return -1;
    }
    return 0;
}

static int xenstore_device_dir(struct xenvm *xen,
			       char *ftype, char *btype, int vdev,
			       char *fe, char *be, int len)
{
    snprintf(fe, len, "/local/domain/%d/device/%s/%d",
	     xen->domid, ftype, vdev);
    snprintf(be, len, "/local/domain/0/backend/%s/%d/%d",
	     btype, xen->domid, vdev);

    xenstore_device_mkdir(xen, fe, XS_PERM_WRITE);
    xenstore_device_mkdir(xen, be, XS_PERM_READ);
    return 0;
}

static int xenstore_device_all(struct xenvm *xen, char *fe, char *be)
{
    /* frontend */
    xenstore_device_set_str(xen, fe, "backend",         be);
    xenstore_device_set_int(xen, fe, "backend-id",      0);
    xenstore_device_set_str(xen, fe, "protocol",
			    xenstore_device_protocol(xen));
    xenstore_device_set_int(xen, fe, "state",           XenbusStateInitialising);

    /* backend */
    xenstore_device_set_str(xen, be, "frontend",        fe);
    xenstore_device_set_int(xen, be, "frontend-id",     xen->domid);
    xenstore_device_set_str(xen, be, "domain",          xen->name);
    xenstore_device_set_int(xen, be, "online",          1);
    xenstore_device_set_int(xen, be, "state",           XenbusStateInitialising);

    return 0;
}

static int xenstore_device_blk(struct xenvm *xen, int vdev, char *dev,
			       char *file, char *devtype, char *mode)
{
    char fe[256], be[256];

    xenstore_device_dir(xen, "vbd", "blkbackd", vdev, fe, be, sizeof(fe));

    /* frontend */
    xenstore_device_set_int(xen, fe, "virtual-device",  vdev);
    xenstore_device_set_str(xen, fe, "device-type",     devtype);

    /* backend */
    xenstore_device_set_str(xen, be, "dev",             dev);
    xenstore_device_set_str(xen, be, "type",            "file");
    xenstore_device_set_str(xen, be, "params",          file);
    xenstore_device_set_str(xen, be, "mode",            mode);
/*
/local/domain/0/backend/vbd/3/51712/physical-device              :  "fd:7"
/local/domain/0/backend/vbd/3/51712/hotplug-status               :  "connected"
*/
    /* common stuff */
    return xenstore_device_all(xen, fe, be);
}

static int xenstore_device_vfb(struct xenvm *xen, int vdev, char *type)
{
    char fe[256], be[256];
    
    xenstore_device_dir(xen, "vfb", "vfb", vdev, fe, be, sizeof(fe));

    /* backend */
    xenstore_device_set_str(xen, be, "type",            type);
/*
/local/domain/0/backend/vfb/2/0/xauthority                       :  "/root/.Xauthority"
/local/domain/0/backend/vfb/2/0/display                          :  "localhost:10.0"
*/

    /* common stuff */
    return xenstore_device_all(xen, fe, be);
}

static int xenstore_device_vkbd(struct xenvm *xen, int vdev)
{
    char fe[256], be[256];

    xenstore_device_dir(xen, "vkbd", "vkbd", vdev, fe, be, sizeof(fe));
    return xenstore_device_all(xen, fe, be);
}

static int xenstore_setup(struct xenvm *xen)
{
    xenstore_setup_host(xen);

    xenstore_vm_set_str(xen, "image/ostype",  "linux");
    xenstore_vm_set_str(xen, "image/kernel",  xen->kernel);
    if (xen->ramdisk)
	xenstore_vm_set_str(xen, "image/ramdisk", xen->ramdisk);
    if (xen->cmdline)
	xenstore_vm_set_str(xen, "image/cmdline", xen->cmdline);

    xenstore_vm_set_str(xen, "name",          xen->name);
    xenstore_vm_set_str(xen, "uuid",          xen->vm+4);
    xenstore_vm_set_int(xen, "vcpu_avail",    1);
    xenstore_vm_set_int(xen, "vcpus",         1);
    xenstore_vm_set_int(xen, "memory",        PG_TO_MB(xen->pg_guest));
    xenstore_vm_set_int(xen, "maxmem",        PG_TO_MB(xen->pg_guest));
    
    xenstore_domain_set_str(xen, "name",               xen->name);
    xenstore_domain_set_int(xen, "domid",              xen->domid);
    xenstore_domain_set_str(xen, "vm",                 xen->vm);

    xenstore_domain_set_int(xen, "memory/target",      xen->pg_guest * 4);

    xenstore_domain_set_int(xen, "console/ring-ref",
			    xen->console_pfn + xen->mfn_guest);
    xenstore_domain_set_int(xen, "console/port",
			    xen->console_event);

    xenstore_domain_set_int(xen, "store/ring-ref",
			    xen->xenstore_pfn + xen->mfn_guest);
    xenstore_domain_set_int(xen, "store/port",
			    xen->xenstore_event);

    xenstore_domain_permit(xen);

    if (xen->disk)
	xenstore_device_blk(xen, 202 * 256 + 0, "xvda", xen->disk, "disk", "w");
    if (xen->cdrom)
	xenstore_device_blk(xen, 202 * 256 + 16, "xvdb", xen->cdrom, "cdrom", "r");
    if (xen->vfb) {
	xenstore_device_vfb(xen, 0, "vnc");
	xenstore_device_vkbd(xen, 0);
	xenstore_vm_set_str(xen, "vncpassword", "");
    }

    xs_introduce_domain(xen->xenstore,
			xen->domid,
			xen->xenstore_pfn + xen->mfn_guest,
			xen->xenstore_event);
    return 0;
}

static int xenstore_cleanup_be(struct xenvm *xen, char *btype)
{
    char abspath[256];

    snprintf(abspath, sizeof(abspath), "/local/domain/0/backend/%s/%d",
	     btype, xen->domid);
    return xs_rm(xen->xenstore, 0, abspath);
}

static int xenstore_cleanup(struct xenvm *xen)
{
    char abspath[256];

    /* domain */
    xs_release_domain(xen->xenstore, xen->domid);
    snprintf(abspath, sizeof(abspath), "/local/domain/%d", xen->domid);
    xs_rm(xen->xenstore, 0, abspath);
    xs_rm(xen->xenstore, 0, xen->vm);

    /* backends */
    if (xen->disk || xen->cdrom) {
	xenstore_cleanup_be(xen, "blkbackd");
    }
    if (xen->vfb) {
	xenstore_cleanup_be(xen, "vfb");
	xenstore_cleanup_be(xen, "vkbd");
    }
    return 0;
}

/* ------------------------------------------------------------------ */

static int xenstore_event(struct xenvm *xen, int port, void *data)
{
    if (!xen->xenstore) {
	xenstore_event_internal(xen);
    } else {
	evtchn_notify(xen, xen->xenstore_event);
    }
    return 0;
}

/* before building domain */
int xenstore_init_early(struct xenvm *xen)
{
    xen->xenstore = xs_daemon_open();
    if (NULL == xen->xenstore) {
	d0printf("%s: can't connect to %s (xenstored not running?)\n",
		 __FUNCTION__, xs_daemon_socket());
	return -1;
    }

    xen->xenstore_event = evtchn_port(xen, "xenstore");
    xenner_register_cb(xen, XENNER_CB_EVTCHN, "xenstore", xen->xenstore_event,
		       NULL, xenstore_event);
    return 0;
}

/* after building domain (shared pages known) */
int xenstore_init_late(struct xenvm *xen)
{
    if (!xen->xenstore)
	return -1;
    xenstore_setup(xen);
    return 0;
}

void xenstore_fini(struct xenvm *xen)
{
    if (!xen->xenstore)
	return;
    xenstore_cleanup(xen);
    xs_daemon_close(xen->xenstore);
}
